"""Contains a frame reader and writer classes."""

from __future__ import annotations

import asyncio
from asyncio import IncompleteReadError, StreamReader, StreamWriter
import logging
from typing import Final, NamedTuple

from pyplumio.const import DeviceType
from pyplumio.devices import is_known_device_type
from pyplumio.exceptions import ChecksumError, ReadError, UnknownDeviceError
from pyplumio.frames import (
    DELIMITER_SIZE,
    FRAME_START,
    HEADER_SIZE,
    Frame,
    bcc,
    struct_header,
)
from pyplumio.helpers.timeout import timeout

READER_TIMEOUT: Final = 10
WRITER_TIMEOUT: Final = 10

MIN_FRAME_LENGTH: Final = 10
MAX_FRAME_LENGTH: Final = 1000

_LOGGER = logging.getLogger(__name__)


class FrameWriter:
    """Represents a frame writer."""

    __slots__ = ("_writer",)

    _writer: StreamWriter

    def __init__(self, writer: StreamWriter) -> None:
        """Initialize a new frame writer."""
        self._writer = writer

    @timeout(WRITER_TIMEOUT)
    async def write(self, frame: Frame) -> None:
        """Send the frame and wait until send buffer is empty."""
        self._writer.write(frame.bytes)
        await self._writer.drain()
        _LOGGER.debug("Sent frame: %s, bytes: %s", frame, frame.bytes)

    async def close(self) -> None:
        """Close the frame writer."""
        try:
            self._writer.close()
            await self.wait_closed()
        except (OSError, asyncio.TimeoutError):
            _LOGGER.exception(
                "Failed to close the frame writer due to an unexpected error"
            )

    @timeout(WRITER_TIMEOUT)
    async def wait_closed(self) -> None:
        """Wait until the frame writer is closed."""
        await self._writer.wait_closed()


class Header(NamedTuple):
    """Represents a frame header."""

    frame_length: int
    recipient: int
    sender: int
    econet_type: int
    econet_version: int


class FrameReader:
    """Represents a frame reader."""

    __slots__ = ("_reader",)

    _reader: StreamReader

    def __init__(self, reader: StreamReader) -> None:
        """Initialize a new frame reader."""
        self._reader = reader

    async def _read_header(self) -> tuple[Header, bytes]:
        """Locate and read a frame header.

        Raise pyplumio.ReadError if header size is too small and
        OSError if serial connection is broken.
        """
        while buffer := await self._reader.read(DELIMITER_SIZE):
            if FRAME_START not in buffer:
                continue

            try:
                buffer += await self._reader.readexactly(HEADER_SIZE - DELIMITER_SIZE)
            except IncompleteReadError as e:
                raise ReadError(
                    f"Incomplete header, expected {e.expected} bytes"
                ) from e

            return Header(*struct_header.unpack_from(buffer)[DELIMITER_SIZE:]), buffer

        raise OSError("Serial connection broken")

    @timeout(READER_TIMEOUT)
    async def read(self) -> Frame | None:
        """Read the frame and return corresponding handler object.

        Raise pyplumio.UnknownDeviceError when sender device has an
        unknown address, raise pyplumio.ReadError on unexpected frame
        length or incomplete frame, raise pyplumio.ChecksumError on
        incorrect frame checksum.
        """
        header, buffer = await self._read_header()
        frame_length, recipient, sender, econet_type, econet_version = header

        if recipient not in (DeviceType.ECONET, DeviceType.ALL):
            # Not an intended recipient, ignore the frame.
            return None

        if not is_known_device_type(sender):
            raise UnknownDeviceError(f"Unknown sender type ({sender})")

        if frame_length > MAX_FRAME_LENGTH or frame_length < MIN_FRAME_LENGTH:
            raise ReadError(
                f"Unexpected frame length ({frame_length}), expected between "
                f"{MIN_FRAME_LENGTH} and {MAX_FRAME_LENGTH}"
            )

        try:
            buffer += await self._reader.readexactly(frame_length - HEADER_SIZE)
        except IncompleteReadError as e:
            raise ReadError(f"Incomplete frame, expected {e.expected} bytes") from e

        if (checksum := bcc(buffer[:-2])) and checksum != buffer[-2]:
            raise ChecksumError(
                f"Incorrect frame checksum: calculated {checksum}, "
                f"expected {buffer[-2]}. "
                f"Frame data: {buffer.hex()}"
            )

        frame = await Frame.create(
            frame_type=buffer[HEADER_SIZE],
            recipient=DeviceType(recipient),
            sender=DeviceType(sender),
            econet_type=econet_type,
            econet_version=econet_version,
            message=buffer[HEADER_SIZE + 1 : -2],
        )
        _LOGGER.debug("Received frame: %s, bytes: %s", frame, buffer.hex())

        return frame



"""Contains a test suite."""

import importlib
import json
import os
import pathlib
from typing import Any, Final

import pytest

TESTDATA_DIR: Final = "testdata"


def load_and_create_class_instance(module_name: str, class_name: str, **kwargs):
    """Load a class and creates it's instance."""
    return getattr(importlib.import_module(module_name), class_name)(**kwargs)


def try_int(key: Any) -> Any:
    """Try to convert key to integer or return key unchanged on error."""
    try:
        return int(key)
    except ValueError:
        return key


def decode_hinted_objects(d: Any) -> Any:
    """Decode a hinted JSON objects."""
    if "__module__" in d and "__class__" in d:
        module_name = d.pop("__module__")
        class_name = d.pop("__class__")
        return load_and_create_class_instance(module_name, class_name, **d)

    if "__bytearray__" in d:
        return bytearray.fromhex("".join(d["items"]))

    if "__tuple__" in d:
        return tuple(d["items"])

    return {try_int(k): v for k, v in d.items()}


def load_json_test_data(path: str):
    """Load test data from JSON file."""
    abs_path = "/".join([os.path.dirname(__file__), TESTDATA_DIR, path])
    file = pathlib.Path(abs_path)
    with open(file, encoding="utf-8") as fp:
        return json.load(fp, object_hook=decode_hinted_objects)


def load_json_parameters(path: str):
    """Prepare JSON test data for parametrization."""
    test_data = load_json_test_data(path)
    return [pytest.param(x["message"], x["data"], id=x["id"]) for x in test_data]



"""Contains fuel consumption structure decoder."""

from __future__ import annotations

import math
from typing import Any, Final

from pyplumio.helpers.data_types import Float
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_FUEL_CONSUMPTION: Final = "fuel_consumption"


class FuelConsumptionStructure(StructureDecoder):
    """Represents a fuel consumption sensor data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        fuel_consumption = Float.from_bytes(message, offset)
        offset += fuel_consumption.size

        if math.isnan(fuel_consumption.value):
            return ensure_dict(data), offset

        return (
            ensure_dict(data, {ATTR_FUEL_CONSUMPTION: fuel_consumption.value}),
            offset,
        )



"""Contains tests for the init module."""

from typing import Final

from pyplumio import (
    SerialConnection,
    TcpConnection,
    open_serial_connection,
    open_tcp_connection,
)

DEVICE: Final = "/dev/ttyUSB0"
IP: Final = "1.1.1.1"


def test_connection_helpers() -> None:
    """Test open connection helpers."""
    serial_connection = open_serial_connection(DEVICE, baudrate=9600)
    assert isinstance(serial_connection, SerialConnection)
    assert serial_connection.device == DEVICE
    assert serial_connection.baudrate == 9600
    tcp_connection = open_tcp_connection(IP, port=3939)
    assert isinstance(tcp_connection, TcpConnection)
    assert tcp_connection.host == IP
    assert tcp_connection.port == 3939



"""Contains a temperatures structure decoder."""

from __future__ import annotations

import math
from typing import Any, Final

from pyplumio.helpers.data_types import Float
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_HEATING_TEMP: Final = "heating_temp"
ATTR_FEEDER_TEMP: Final = "feeder_temp"
ATTR_WATER_HEATER_TEMP: Final = "water_heater_temp"
ATTR_OUTSIDE_TEMP: Final = "outside_temp"
ATTR_RETURN_TEMP: Final = "return_temp"
ATTR_EXHAUST_TEMP: Final = "exhaust_temp"
ATTR_OPTICAL_TEMP: Final = "optical_temp"
ATTR_UPPER_BUFFER_TEMP: Final = "upper_buffer_temp"
ATTR_LOWER_BUFFER_TEMP: Final = "lower_buffer_temp"
ATTR_UPPER_SOLAR_TEMP: Final = "upper_solar_temp"
ATTR_LOWER_SOLAR_TEMP: Final = "lower_solar_temp"
ATTR_FIREPLACE_TEMP: Final = "fireplace_temp"
ATTR_TOTAL_GAIN: Final = "total_gain"
ATTR_HYDRAULIC_COUPLER_TEMP: Final = "hydraulic_coupler_temp"
ATTR_EXCHANGER_TEMP: Final = "exchanger_temp"
ATTR_AIR_IN_TEMP: Final = "air_in_temp"
ATTR_AIR_OUT_TEMP: Final = "air_out_temp"

TEMPERATURES: tuple[str, ...] = (
    ATTR_HEATING_TEMP,
    ATTR_FEEDER_TEMP,
    ATTR_WATER_HEATER_TEMP,
    ATTR_OUTSIDE_TEMP,
    ATTR_RETURN_TEMP,
    ATTR_EXHAUST_TEMP,
    ATTR_OPTICAL_TEMP,
    ATTR_UPPER_BUFFER_TEMP,
    ATTR_LOWER_BUFFER_TEMP,
    ATTR_UPPER_SOLAR_TEMP,
    ATTR_LOWER_SOLAR_TEMP,
    ATTR_FIREPLACE_TEMP,
    ATTR_TOTAL_GAIN,
    ATTR_HYDRAULIC_COUPLER_TEMP,
    ATTR_EXCHANGER_TEMP,
    ATTR_AIR_IN_TEMP,
    ATTR_AIR_OUT_TEMP,
)


class TemperaturesStructure(StructureDecoder):
    """Represents a temperatures data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        data = ensure_dict(data)
        temperatures = message[offset]
        offset += 1
        for _ in range(temperatures):
            index = message[offset]
            offset += 1
            temp = Float.from_bytes(message, offset)
            offset += temp.size
            if (not math.isnan(temp.value)) and 0 <= index < len(TEMPERATURES):
                # Temperature exists and index is in the correct range.
                data[TEMPERATURES[index]] = temp.value

        return data, offset



"""Contains a schedule helper classes."""

from __future__ import annotations

from collections.abc import Iterable, Iterator, MutableMapping
from dataclasses import dataclass
import datetime as dt
from functools import lru_cache
from typing import Annotated, Final, get_args

from pyplumio.const import STATE_OFF, STATE_ON, FrameType, State
from pyplumio.devices import PhysicalDevice
from pyplumio.frames import Request
from pyplumio.structures.schedules import collect_schedule_data

TIME_FORMAT: Final = "%H:%M"

MIDNIGHT: Final = "00:00"
MIDNIGHT_DT = dt.datetime.strptime(MIDNIGHT, TIME_FORMAT)

STEP = dt.timedelta(minutes=30)

Time = Annotated[str, "Time string in %H:%M format"]


def _get_time(
    index: int, start: dt.datetime = MIDNIGHT_DT, step: dt.timedelta = STEP
) -> Time:
    """Return time for a specific index."""
    time_dt = start + (step * index)
    return time_dt.strftime(TIME_FORMAT)


@lru_cache(maxsize=10)
def _get_time_range(start: Time, end: Time, step: dt.timedelta = STEP) -> list[Time]:
    """Get a time range.

    Start and end boundaries should be specified in %H:%M format.
    Both are inclusive.
    """
    start_dt = dt.datetime.strptime(start, TIME_FORMAT)
    end_dt = dt.datetime.strptime(end, TIME_FORMAT)

    if end_dt == MIDNIGHT_DT:
        # Upper boundary of the interval is midnight.
        end_dt += dt.timedelta(hours=24) - step

    if end_dt <= start_dt:
        raise ValueError(
            f"Invalid time range: start time ({start}) must be earlier "
            f"than end time ({end})."
        )

    seconds = (end_dt - start_dt).total_seconds()
    steps = seconds // step.total_seconds() + 1

    return [_get_time(index, start=start_dt, step=step) for index in range(int(steps))]


class ScheduleDay(MutableMapping):
    """Represents a single day of schedule."""

    __slots__ = ("_schedule",)

    _schedule: dict[Time, bool]

    def __init__(self, schedule: dict[Time, bool]) -> None:
        """Initialize a new schedule day."""
        self._schedule = schedule

    def __repr__(self) -> str:
        """Return serializable representation of the class."""
        return f"ScheduleDay({self._schedule})"

    def __len__(self) -> int:
        """Return a schedule length."""
        return self._schedule.__len__()

    def __iter__(self) -> Iterator[Time]:
        """Return an iterator."""
        return self._schedule.__iter__()

    def __getitem__(self, time: Time) -> State:
        """Return a schedule item."""
        state = self._schedule.__getitem__(time)
        return STATE_ON if state else STATE_OFF

    def __delitem__(self, time: Time) -> None:
        """Delete a schedule item."""
        self._schedule.__delitem__(time)

    def __setitem__(self, time: Time, state: State | bool) -> None:
        """Set a schedule item."""
        if state in get_args(State):
            state = True if state == STATE_ON else False
        if isinstance(state, bool):
            self._schedule.__setitem__(time, state)
        else:
            raise TypeError(
                f"Expected boolean value or one of: {', '.join(get_args(State))}."
            )

    def set_state(
        self, state: State | bool, start: Time = MIDNIGHT, end: Time = MIDNIGHT
    ) -> None:
        """Set a schedule interval state."""
        for time in _get_time_range(start, end):
            self.__setitem__(time, state)

    def set_on(self, start: Time = MIDNIGHT, end: Time = MIDNIGHT) -> None:
        """Set a schedule interval state to 'on'."""
        self.set_state(STATE_ON, start, end)

    def set_off(self, start: Time = MIDNIGHT, end: Time = MIDNIGHT) -> None:
        """Set a schedule interval state to 'off'."""
        self.set_state(STATE_OFF, start, end)

    @property
    def schedule(self) -> dict[Time, bool]:
        """Return the schedule."""
        return self._schedule

    @classmethod
    def from_iterable(cls: type[ScheduleDay], intervals: Iterable[bool]) -> ScheduleDay:
        """Make schedule day from iterable."""
        return cls({_get_time(index): state for index, state in enumerate(intervals)})


@dataclass
class Schedule(Iterable):
    """Represents a weekly schedule."""

    __slots__ = (
        "name",
        "device",
        "sunday",
        "monday",
        "tuesday",
        "wednesday",
        "thursday",
        "friday",
        "saturday",
    )

    name: str
    device: PhysicalDevice

    sunday: ScheduleDay
    monday: ScheduleDay
    tuesday: ScheduleDay
    wednesday: ScheduleDay
    thursday: ScheduleDay
    friday: ScheduleDay
    saturday: ScheduleDay

    def __iter__(self) -> Iterator[ScheduleDay]:
        """Return list of days."""
        return (
            self.sunday,
            self.monday,
            self.tuesday,
            self.wednesday,
            self.thursday,
            self.friday,
            self.saturday,
        ).__iter__()

    async def commit(self) -> None:
        """Commit a weekly schedule to the device."""
        await self.device.queue.put(
            await Request.create(
                FrameType.REQUEST_SET_SCHEDULE,
                recipient=self.device.address,
                data=collect_schedule_data(self.name, self.device),
            )
        )



"""Contains tests for the object factory."""

import pytest

from pyplumio.frames import Request
from pyplumio.frames.requests import StopMasterRequest
from pyplumio.helpers.factory import create_instance


async def test_get_object() -> None:
    """Test getting an object via class path."""
    cls = await create_instance("frames.requests.StopMasterRequest", cls=Request)
    assert isinstance(cls, StopMasterRequest)


async def test_get_object_with_incorrect_base_class() -> None:
    """Test getting an object via class path with incorrect base class."""
    with pytest.raises(TypeError) as exc_info:
        await create_instance("frames.responses.UIDResponse", cls=Request)

    assert str(exc_info.value) == (
        "Expected instance of 'Request', but got 'UIDResponse' from 'UIDResponse'"
    )


async def test_get_object_with_nonexistent_class() -> None:
    """Test getting an object via class path for nonexistent class."""
    with pytest.raises(AttributeError):
        await create_instance("frames.requests.NonExistent", cls=Request)


async def test_get_object_with_nonexistent_module() -> None:
    """Test getting an object via class path for a class within nonexistent module."""
    with pytest.raises(ModuleNotFoundError):
        await create_instance("frames.request.StopMasterRequest", cls=Request)



"""Contains tests for the event manager."""

from typing import Any
from unittest.mock import AsyncMock, call, patch

import pytest

from pyplumio.helpers.event_manager import EventManager


@pytest.fixture(name="event_manager")
def fixture_event_manager() -> EventManager:
    """Return an event manager object."""
    event_manager = EventManager[Any]()
    event_manager.data = {"test_key": "test_value"}
    return event_manager


def test_getattr(event_manager: EventManager) -> None:
    """Test getting an event manager attribute."""
    assert event_manager.test_key == "test_value"


async def test_wait_for(event_manager: EventManager) -> None:
    """Test waiting for event."""
    with (
        patch("asyncio.wait_for") as mock_wait_for,
        patch(
            "pyplumio.helpers.event_manager.EventManager.create_event"
        ) as mock_create_event,
    ):
        mock_create_event.wait = AsyncMock()
        await event_manager.wait_for("test_key2")

    mock_create_event.assert_called_with("test_key2")
    mock_wait_for.assert_awaited_once_with(
        mock_create_event.return_value.wait.return_value, timeout=None
    )


async def test_get(event_manager: EventManager) -> None:
    """Test getting an event value."""
    assert await event_manager.get("test_key") == "test_value"
    assert event_manager.test_key == "test_value"
    with pytest.raises(AttributeError):
        event_manager.nonexistent


def test_get_nowait(event_manager: EventManager) -> None:
    """Test getting an event value without waiting."""
    assert event_manager.get_nowait("test_key") == "test_value"
    assert event_manager.get_nowait("test_key2") is None


async def test_load(event_manager: EventManager) -> None:
    """Test loading event data."""
    callback = AsyncMock(return_value=True)
    callback2 = AsyncMock(return_value=True)
    event_manager.subscribe("test_key1", callback)
    event_manager.subscribe("test_key2", callback2)
    event_manager.load_nowait({"test_key2": "test_value2"})
    await event_manager.wait_until_done()
    callback.assert_not_awaited()
    callback2.assert_awaited_once_with("test_value2")


async def test_load_nowait(event_manager: EventManager) -> None:
    """Test loading event data without waiting."""
    callback = AsyncMock(return_value=True)
    callback2 = AsyncMock(return_value=True)
    event_manager.subscribe("test_key1", callback)
    event_manager.subscribe("test_key2", callback2)
    await event_manager.load({"test_key2": "test_value2"})
    callback.assert_not_awaited()
    callback2.assert_awaited_once_with("test_value2")


async def test_subscribe(event_manager: EventManager) -> None:
    """Test subscribing to an event."""
    callback = AsyncMock(return_value=True)
    assert event_manager.subscribe("test_key2", callback) is callback
    event_manager.dispatch_nowait("test_key2", "test_value2")
    event_manager.dispatch_nowait("test_key2", "test_value3")
    await event_manager.wait_until_done()
    callback.assert_has_awaits([call("test_value2"), call("test_value3")])


async def test_subscribe_once(event_manager: EventManager) -> None:
    """Test subscribing to an event once."""
    callback = AsyncMock(return_value=True)
    event_manager.subscribe_once("test_key2", callback)
    event_manager.dispatch_nowait("test_key2", "test_value2")
    event_manager.dispatch_nowait("test_key2", "test_value3")
    await event_manager.wait_until_done()
    callback.assert_awaited_once_with("test_value2")
    callback.reset_mock()

    # Test getting the reference and unsubscribing.
    with patch(
        "pyplumio.helpers.event_manager.EventManager.unsubscribe"
    ) as mock_unsubscribe:
        callback2 = event_manager.subscribe_once("test_key3", callback)
        event_manager.unsubscribe("test_key3", callback2)
        await event_manager.wait_until_done()

    mock_unsubscribe.assert_called_once_with("test_key3", callback2)
    callback.assert_not_awaited()


async def test_subscribe_once_interrupt(event_manager: EventManager) -> None:
    """Test subscribing once doesn't affect other callbacks."""
    callback1 = AsyncMock(return_value=None)
    callback2 = AsyncMock(return_value=None)
    event_manager.subscribe_once("test_key1", callback1)
    assert event_manager.subscribe("test_key1", callback2) is callback2
    event_manager.dispatch_nowait("test_key1", "test_value1")
    event_manager.dispatch_nowait("test_key1", "test_value2")
    await event_manager.wait_until_done()
    callback2.assert_has_awaits([call("test_value1"), call("test_value2")])


async def test_unsubscribe(event_manager: EventManager) -> None:
    """Test unsubscribing from the event."""
    callback = AsyncMock(return_value=True)
    event_manager.subscribe("test_key2", callback)
    assert event_manager.unsubscribe("test_key2", callback)
    assert not event_manager.unsubscribe("test_key3", callback)
    event_manager.dispatch_nowait("test_key2", "test_value2")
    await event_manager.wait_until_done()
    callback.assert_not_awaited()


async def test_create_event(event_manager: EventManager) -> None:
    """Test creating an event."""
    event = event_manager.create_event("test")
    assert event == event_manager.create_event("test")
    assert "test" in event_manager.events
    assert not event_manager.events["test"].is_set()
    event_manager.set_event("test")
    assert event_manager.events["test"].is_set()



"""Contains tests for the utility functions."""

from pyplumio import utils


def test_to_camelcase() -> None:
    """Test string to camelcase converter."""
    assert utils.to_camelcase("make_love_not_war") == "MakeLoveNotWar"
    assert (
        utils.to_camelcase(
            "make_love_not_war", overrides={"love": "LOVE", "not": "not"}
        )
        == "MakeLOVEnotWar"
    )


def test_ensure_dict() -> None:
    """Test ensure dictionary."""
    assert utils.ensure_dict({"foo": "bar"}, {"baz": "foobar"}) == {
        "foo": "bar",
        "baz": "foobar",
    }



"""Contains tests for the schedule helper classes."""

import asyncio
from unittest.mock import Mock, patch

import pytest

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SWITCH,
    ATTR_TYPE,
    STATE_OFF,
    STATE_ON,
    FrameType,
)
from pyplumio.devices import Device, DeviceType
from pyplumio.helpers.schedule import Schedule, ScheduleDay, Time
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETER,
    ATTR_SCHEDULE_SWITCH,
    ATTR_SCHEDULES,
)


@pytest.fixture(name="intervals")
def fixture_intervals() -> list[Time]:
    """Return time intervals for a schedule day."""
    return (
        "00:00, 00:30, 01:00, 01:30, 02:00, 02:30, 03:00, 03:30, 04:00, 04:30, 05:00, "
        "05:30, 06:00, 06:30, 07:00, 07:30, 08:00, 08:30, 09:00, 09:30, 10:00, 10:30, "
        "11:00, 11:30, 12:00, 12:30, 13:00, 13:30, 14:00, 14:30, 15:00, 15:30, 16:00, "
        "16:30, 17:00, 17:30, 18:00, 18:30, 19:00, 19:30, 20:00, 20:30, 21:00, 21:30, "
        "22:00, 22:30, 23:00, 23:30"
    ).split(", ")


@pytest.fixture(name="schedule_day")
def fixture_schedule_day() -> ScheduleDay:
    """Return a schedule day object."""
    schedule_day = ScheduleDay.from_iterable([False for _ in range(48)])
    assert len(schedule_day) == 48
    return schedule_day


@pytest.fixture(name="schedule")
def fixture_schedule(schedule_day: ScheduleDay) -> Schedule:
    """Return a schedule object."""
    return Schedule(
        name="test",
        device=Mock(spec=Device),
        monday=ScheduleDay.from_iterable([True for _ in range(48)]),
        tuesday=schedule_day,
        wednesday=schedule_day,
        thursday=schedule_day,
        friday=schedule_day,
        saturday=schedule_day,
        sunday=schedule_day,
    )


def test_schedule_day(schedule_day: ScheduleDay) -> None:
    """Test a schedule day."""
    schedule_day.set_state(STATE_ON, "00:00", "01:00")
    for time in ("00:00", "00:30", "01:00"):
        assert schedule_day[time] == STATE_ON
    assert schedule_day["01:30"] == schedule_day["02:00"] == STATE_OFF
    schedule_day["02:00"] = STATE_ON
    assert schedule_day.schedule["02:00"]


def test_schedule_day_with_missing_key(schedule_day: ScheduleDay) -> None:
    """Test a schedule day with a missing key."""
    with pytest.raises(KeyError):
        assert schedule_day["00:20"] == STATE_OFF


def test_schedule_day_with_incorrect_interval(schedule_day: ScheduleDay) -> None:
    """Test a schedule day with an incorrect interval."""
    for start, end in (("01:00", "00:30"), ("00:foo", "bar")):
        with pytest.raises(ValueError):
            schedule_day.set_state(STATE_ON, start, end)


def test_schedule_day_with_incorrect_state(schedule_day: ScheduleDay) -> None:
    """Test a schedule day with an incorrect state."""
    with pytest.raises(TypeError):
        schedule_day.set_state("invalid_state", "00:00", "01:00")  # type: ignore[arg-type]


def test_setting_whole_day_schedule(
    schedule_day: ScheduleDay, intervals: list[Time]
) -> None:
    """Test setting schedule for a whole day."""
    schedule_day.set_on()
    assert schedule_day.schedule == dict.fromkeys(intervals, True)
    schedule_day.set_off()
    assert schedule_day.schedule == dict.fromkeys(intervals, False)


def test_setting_schedule_as_iterable(schedule_day: ScheduleDay) -> None:
    """Test setting schedule using iterable methods."""
    schedule_day.set_on("00:30", "01:00")
    schedule_day_iter = iter(schedule_day)
    assert schedule_day[next(schedule_day_iter)] == STATE_OFF
    assert schedule_day[next(schedule_day_iter)] == STATE_ON
    assert "00:30" in schedule_day
    assert schedule_day["00:30"] == STATE_ON
    schedule_day["00:30"] = STATE_OFF
    del schedule_day["00:00"]
    assert len(schedule_day) == 47
    assert "00:00" not in schedule_day


def test_schedule_day_repr(schedule_day: ScheduleDay, intervals: list[Time]) -> None:
    """Test serializable representation of a schedule day."""
    schedule = dict.fromkeys(intervals, False)
    assert repr(schedule_day) == (f"ScheduleDay({schedule})")


def test_schedule(schedule: Schedule) -> None:
    """Test a schedule."""
    schedule_iter = iter(schedule)
    assert next(schedule_iter)["00:00"] == STATE_OFF
    assert next(schedule_iter)["00:00"] == STATE_ON


@patch("pyplumio.helpers.schedule.Request.create")
async def test_schedule_commit(mock_request_create, schedule: Schedule) -> None:
    """Test committing a schedule."""
    schedule.device = Mock(spec=Device)
    schedule.device.address = DeviceType.ECOMAX
    schedule.device.data = {
        f"test_{ATTR_SCHEDULE_SWITCH}": 1,
        f"test_{ATTR_SCHEDULE_PARAMETER}": 2,
        ATTR_SCHEDULES: {"test": schedule},
    }
    schedule.device.queue = Mock(spec=asyncio.Queue)

    await schedule.commit()
    mock_request_create.assert_awaited_once_with(
        FrameType.REQUEST_SET_SCHEDULE,
        recipient=DeviceType.ECOMAX,
        data={
            ATTR_TYPE: "test",
            ATTR_SWITCH: 1,
            ATTR_PARAMETER: 2,
            ATTR_SCHEDULE: schedule,
        },
    )



"""Contains an output flags structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.helpers.data_types import UnsignedInt
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_HEATING_PUMP_FLAG: Final = "heating_pump_flag"
ATTR_WATER_HEATER_PUMP_FLAG: Final = "water_heater_pump_flag"
ATTR_CIRCULATION_PUMP_FLAG: Final = "circulation_pump_flag"
ATTR_SOLAR_PUMP_FLAG: Final = "solar_pump_flag"


class OutputFlagsStructure(StructureDecoder):
    """Represents an output flags structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        output_flags = UnsignedInt.from_bytes(message, offset)
        return (
            ensure_dict(
                data,
                {
                    ATTR_HEATING_PUMP_FLAG: bool(output_flags.value & 0x04),
                    ATTR_WATER_HEATER_PUMP_FLAG: bool(output_flags.value & 0x08),
                    ATTR_CIRCULATION_PUMP_FLAG: bool(output_flags.value & 0x10),
                    ATTR_SOLAR_PUMP_FLAG: bool(output_flags.value & 0x800),
                },
            ),
            offset + output_flags.size,
        )



"""Contains a factory helper."""

from __future__ import annotations

import asyncio
import importlib
import logging
from types import ModuleType
from typing import Any, TypeVar

_LOGGER = logging.getLogger(__name__)

T = TypeVar("T")


async def _import_module(name: str) -> ModuleType:
    """Import module by name."""
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, importlib.import_module, f"pyplumio.{name}")


async def create_instance(class_path: str, /, cls: type[T], **kwargs: Any) -> T:
    """Return a class instance from the class path."""
    module_name, class_name = class_path.rsplit(".", 1)
    try:
        module = await _import_module(module_name)
        instance = getattr(module, class_name)(**kwargs)
        if not isinstance(instance, cls):
            raise TypeError(
                f"Expected instance of '{cls.__name__}', but got "
                f"'{type(instance).__name__}' from '{class_name}'"
            )

        return instance
    except Exception:
        _LOGGER.exception("Failed to create instance for class path '%s'", class_path)
        raise



"""Contains schedule decoder."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from functools import reduce
from typing import Any, Final

from dataslots import dataslots

from pyplumio.const import (
    ATTR_PARAMETER,
    ATTR_SCHEDULE,
    ATTR_SWITCH,
    ATTR_TYPE,
    FrameType,
)
from pyplumio.devices import Device, PhysicalDevice
from pyplumio.exceptions import FrameDataError
from pyplumio.frames import Request
from pyplumio.helpers.parameter import (
    Number,
    NumberDescription,
    Parameter,
    ParameterDescription,
    ParameterValues,
    Switch,
    SwitchDescription,
    unpack_parameter,
)
from pyplumio.structures import Structure
from pyplumio.utils import ensure_dict

ATTR_SCHEDULES: Final = "schedules"
ATTR_SCHEDULE_PARAMETERS: Final = "schedule_parameters"
ATTR_SCHEDULE_SWITCH: Final = "schedule_switch"
ATTR_SCHEDULE_PARAMETER: Final = "schedule_parameter"

SCHEDULE_SIZE: Final = 42  # 6 bytes per day, 7 days total

SCHEDULES: tuple[str, ...] = (
    "heating",
    "water_heater",
    "circulation_pump",
    "boiler_work",
    "boiler_clean",
    "hear_exchanger_clean",
    "mixer_1",
    "mixer_2",
    "mixer_3",
    "mixer_4",
    "mixer_5",
    "mixer_6",
    "mixer_7",
    "mixer_8",
    "mixer_9",
    "mixer_10",
    "thermostat_1",
    "thermostat_2",
    "thermostat_3",
    "circuit_1",
    "circuit_2",
    "circuit_3",
    "circuit_4",
    "circuit_5",
    "circuit_6",
    "circuit_7",
    "panel_1",
    "panel_2",
    "panel_3",
    "panel_4",
    "panel_5",
    "panel_6",
    "panel_7",
    "main_heater_solar",
    "heating_circulation",
    "internal_thermostat",
    "heater",
    "water_heater_2",
    "intake",
    "intake_summer",
)


@dataclass
class ScheduleParameterDescription(ParameterDescription):
    """Represent a schedule parameter description."""

    __slots__ = ()


class ScheduleParameter(Parameter):
    """Represents a schedule parameter."""

    __slots__ = ()

    device: PhysicalDevice
    description: ScheduleParameterDescription

    async def create_request(self) -> Request:
        """Create a request to change the parameter."""
        schedule_name = self.description.name.split("_schedule_", 1)[0]
        return await Request.create(
            FrameType.REQUEST_SET_SCHEDULE,
            recipient=self.device.address,
            data=collect_schedule_data(schedule_name, self.device),
        )


@dataslots
@dataclass
class ScheduleNumberDescription(ScheduleParameterDescription, NumberDescription):
    """Represents a schedule number description."""


class ScheduleNumber(ScheduleParameter, Number):
    """Represents a schedule number."""

    __slots__ = ()

    description: ScheduleNumberDescription


@dataslots
@dataclass
class ScheduleSwitchDescription(ScheduleParameterDescription, SwitchDescription):
    """Represents a schedule switch description."""


class ScheduleSwitch(ScheduleParameter, Switch):
    """Represents a schedule switch."""

    __slots__ = ()

    description: ScheduleSwitchDescription


SCHEDULE_PARAMETERS: list[ScheduleParameterDescription] = [
    description
    for name in SCHEDULES
    for description in [
        ScheduleSwitchDescription(name=f"{name}_{ATTR_SCHEDULE_SWITCH}"),
        ScheduleNumberDescription(name=f"{name}_{ATTR_SCHEDULE_PARAMETER}"),
    ]
]


def collect_schedule_data(name: str, device: Device) -> dict[str, Any]:
    """Return a schedule data collected from the device."""
    return {
        ATTR_TYPE: name,
        ATTR_SWITCH: device.data[f"{name}_{ATTR_SCHEDULE_SWITCH}"],
        ATTR_PARAMETER: device.data[f"{name}_{ATTR_SCHEDULE_PARAMETER}"],
        ATTR_SCHEDULE: device.data[ATTR_SCHEDULES][name],
    }


def _split_byte(byte: int) -> list[bool]:
    """Split single byte into an eight bits."""
    return [bool(byte & (1 << bit)) for bit in reversed(range(8))]


def _join_bits(bits: Sequence[int | bool]) -> int:
    """Join eight bits into a single byte."""
    return reduce(lambda bit, byte: (bit << 1) | byte, bits)


class SchedulesStructure(Structure):
    """Represents a schedule data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def encode(self, data: dict[str, Any]) -> bytearray:
        """Encode data to the bytearray message."""
        try:
            header = bytearray(
                b"\1"
                + SCHEDULES.index(data[ATTR_TYPE]).to_bytes(
                    length=1, byteorder="little"
                )
                + int(data[ATTR_SWITCH]).to_bytes(length=1, byteorder="little")
                + int(data[ATTR_PARAMETER]).to_bytes(length=1, byteorder="little")
            )
            schedule = data[ATTR_SCHEDULE]
        except (KeyError, ValueError) as e:
            raise FrameDataError from e

        return header + bytearray(
            _join_bits(day[i : i + 8])
            for day in schedule
            for i in range(0, len(day), 8)
        )

    def _unpack_schedule(self, message: bytearray) -> list[list[bool]]:
        """Unpack a schedule."""
        offset = self._offset
        schedule = [
            bit
            for i in range(offset, offset + SCHEDULE_SIZE)
            for bit in _split_byte(message[i])
        ]
        self._offset = offset + SCHEDULE_SIZE
        # Split the schedule. Each day consists of 48 half-hour intervals.
        return [schedule[i : i + 48] for i in range(0, len(schedule), 48)]

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        try:
            start = message[offset + 1]
            end = message[offset + 2]
        except IndexError:
            return ensure_dict(data, {ATTR_SCHEDULES: []}), offset

        schedules: list[tuple[int, list[list[bool]]]] = []
        parameters: list[tuple[int, ParameterValues]] = []

        self._offset = offset + 3
        for _ in range(start, start + end):
            index = message[self._offset]
            switch = ParameterValues(
                value=message[self._offset + 1], min_value=0, max_value=1
            )
            parameter = unpack_parameter(message, self._offset + 2)
            self._offset += 5
            schedules.append((index, self._unpack_schedule(message)))
            parameters.append((index * 2, switch))
            if parameter is not None:
                parameters.append((index * 2 + 1, parameter))

        return (
            ensure_dict(
                data, {ATTR_SCHEDULES: schedules, ATTR_SCHEDULE_PARAMETERS: parameters}
            ),
            self._offset,
        )



"""Contains a mixer sensors structure decoder."""

from __future__ import annotations

from collections.abc import Generator
import math
from typing import Any, Final

from pyplumio.const import ATTR_CURRENT_TEMP, ATTR_TARGET_TEMP
from pyplumio.helpers.data_types import Float
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_PUMP: Final = "pump"
ATTR_MIXERS_AVAILABLE: Final = "mixers_available"
ATTR_MIXERS_CONNECTED: Final = "mixers_connected"
ATTR_MIXER_SENSORS: Final = "mixer_sensors"

MIXER_SENSOR_SIZE: Final = 8


class MixerSensorsStructure(StructureDecoder):
    """Represents a mixer sensors data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _unpack_mixer_sensors(self, message: bytearray) -> dict[str, Any] | None:
        """Unpack sensors for a mixer."""
        offset = self._offset
        current_temp = Float.from_bytes(message, offset)
        try:
            return (
                {
                    ATTR_CURRENT_TEMP: current_temp.value,
                    ATTR_TARGET_TEMP: message[offset + 4],
                    ATTR_PUMP: bool(message[offset + 6] & 0x01),
                }
                if not math.isnan(current_temp.value)
                else None
            )
        finally:
            self._offset = offset + MIXER_SENSOR_SIZE

    def _mixer_sensors(
        self, message: bytearray, mixers: int
    ) -> Generator[tuple[int, dict[str, Any]], None, None]:
        """Get sensors for a mixer."""
        for index in range(mixers):
            if sensors := self._unpack_mixer_sensors(message):
                yield (index, sensors)

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        mixers = message[offset]
        self._offset = offset + 1
        mixer_sensors = dict(self._mixer_sensors(message, mixers))
        return (
            ensure_dict(
                data,
                {
                    ATTR_MIXER_SENSORS: mixer_sensors,
                    ATTR_MIXERS_AVAILABLE: mixers,
                    ATTR_MIXERS_CONNECTED: len(mixer_sensors),
                },
            ),
            self._offset,
        )



"""Contains a frame versions structure decoder."""

from __future__ import annotations

from contextlib import suppress
from typing import Any, Final

from pyplumio.const import FrameType
from pyplumio.helpers.data_types import UnsignedShort
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_FRAME_VERSIONS: Final = "frame_versions"


class FrameVersionsStructure(StructureDecoder):
    """Represents a frame version data structure."""

    __slots__ = ("_offset",)

    _offset: int

    def _unpack_frame_versions(self, message: bytearray) -> tuple[FrameType | int, int]:
        """Unpack frame versions."""
        frame_type = message[self._offset]
        version = UnsignedShort.from_bytes(message, self._offset + 1)
        self._offset += version.size + 1
        with suppress(ValueError):
            frame_type = FrameType(frame_type)

        return frame_type, version.value

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        frame_versions = message[offset]
        self._offset = offset + 1
        return (
            ensure_dict(
                data,
                {
                    ATTR_FRAME_VERSIONS: dict(
                        self._unpack_frame_versions(message)
                        for _ in range(frame_versions)
                    )
                },
            ),
            self._offset,
        )



"""Contains tests for the data type helper classes."""

import pytest

from pyplumio.helpers import data_types


@pytest.mark.parametrize(
    ("one", "another"),
    [
        (
            data_types.SignedChar.from_bytes(bytearray([0x16])),
            data_types.SignedChar.from_bytes(bytearray([0x16])),
        ),
        (
            data_types.SignedChar.from_bytes(bytearray([0x16])),
            22,
        ),
        (
            data_types.SignedChar(),
            data_types.SignedChar(),
        ),
        (
            data_types.SignedChar(),
            "not_implemented",
        ),
    ],
)
def test_type_eq(one, another):
    """Test a generic type comparison."""
    if another != "not_implemented":
        assert one == another
    else:
        assert one.__eq__(another) is NotImplemented


def test_undefined() -> None:
    """Test an undefined data type."""
    buffer = bytearray()
    data_type = data_types.Undefined.from_bytes(buffer)
    assert data_type.value is None
    assert data_type.size == 0
    assert not data_type.to_bytes()
    assert repr(data_type) == "Undefined(value=None)"
    assert repr(data_types.Undefined()) == "Undefined()"
    assert data_type == data_types.Undefined.from_bytes(buffer)


def test_signed_char() -> None:
    """Test a signed char data type."""
    buffer = bytearray([0x16])
    data_type = data_types.SignedChar.from_bytes(buffer)
    assert data_type.value == 22
    assert data_type.size == 1
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "SignedChar(value=22)"
    assert repr(data_types.SignedChar()) == "SignedChar()"
    assert data_type == data_types.SignedChar.from_bytes(buffer)
    assert data_type == 22


def test_short() -> None:
    """Test a short data type."""
    buffer = bytearray([0xEC, 0xFF])
    data_type = data_types.Short.from_bytes(buffer)
    assert data_type.value == -20
    assert data_type.size == 2
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "Short(value=-20)"
    assert repr(data_types.Short()) == "Short()"
    assert data_type == data_types.Short.from_bytes(buffer)
    assert data_type == -20


def test_int() -> None:
    """Test an integer data type."""
    buffer = bytearray([0x01, 0x9A, 0xFF, 0xFF])
    data_type = data_types.Int.from_bytes(buffer)
    assert data_type.value == -26111
    assert data_type.size == 4
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "Int(value=-26111)"
    assert repr(data_types.Int()) == "Int()"
    assert data_type == data_types.Int.from_bytes(buffer)
    assert data_type == -26111


def test_unsigned_char() -> None:
    """Test an signed char data type."""
    buffer = bytearray([0x3])
    data_type = data_types.UnsignedChar.from_bytes(buffer)
    assert data_type.value == 3
    assert data_type.size == 1
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "UnsignedChar(value=3)"
    assert repr(data_types.UnsignedChar()) == "UnsignedChar()"
    assert data_type == data_types.UnsignedChar.from_bytes(buffer)
    assert data_type == 3


def test_ushort() -> None:
    """Test an unsigned short data type."""
    buffer = bytearray([0x2A, 0x01])
    data_type = data_types.UnsignedShort.from_bytes(buffer)
    assert data_type.value == 298
    assert data_type.size == 2
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "UnsignedShort(value=298)"
    assert repr(data_types.UnsignedShort()) == "UnsignedShort()"
    assert data_type == data_types.UnsignedShort.from_bytes(buffer)
    assert data_type == 298


def test_uint() -> None:
    """Test an unsigned integer data type."""
    buffer = bytearray([0x9A, 0x3F, 0x0, 0x0])
    data_type = data_types.UnsignedInt.from_bytes(buffer)
    assert data_type.value == 16282
    assert data_type.size == 4
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "UnsignedInt(value=16282)"
    assert repr(data_types.UnsignedInt()) == "UnsignedInt()"
    assert data_type == data_types.UnsignedInt.from_bytes(buffer)
    assert data_type == 16282


def test_float() -> None:
    """Test a float data type."""
    buffer = bytearray([0x0, 0x0, 0x40, 0x41])
    data_type = data_types.Float.from_bytes(buffer)
    assert data_type.value == 12.0
    assert data_type.size == 4
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "Float(value=12.0)"
    assert repr(data_types.Float()) == "Float()"
    assert data_type == data_types.Float.from_bytes(buffer)
    assert data_type == 12.0


def test_double() -> None:
    """Test a double data type."""
    buffer = bytearray([0x3D, 0x0A, 0xD7, 0xA3, 0x70, 0x3D, 0x28, 0x40])
    data_type = data_types.Double.from_bytes(buffer)
    assert data_type.value == 12.12
    assert data_type.size == 8
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "Double(value=12.12)"
    assert repr(data_types.Double()) == "Double()"
    assert data_type == data_types.Double.from_bytes(buffer)
    assert data_type == 12.12


def test_bitarray() -> None:
    """Test a bit array data type."""
    buffer = bytearray([0x55])
    data_type = data_types.BitArray.from_bytes(buffer)
    for index, value in enumerate([1, 0, 1, 0, 1, 0, 1, 0]):
        next_bit = data_type.next(index)
        assert data_type.value == bool(value)
        if index < 7:
            assert next_bit == index + 1
            assert data_type.size == 0
        else:
            assert next_bit == 0
            assert data_type.size == 1

    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "BitArray(value=85, index=7)"
    assert repr(data_types.BitArray()) == "BitArray(index=0)"
    assert data_type == data_types.BitArray.from_bytes(buffer)
    assert data_type == 85
    assert data_types.BitArray().pack() == b""


def test_bitarray_no_value() -> None:
    """Test a bit array data type with no value."""
    data_type = data_types.BitArray()
    with pytest.raises(ValueError):
        assert not data_type.value


def test_int64() -> None:
    """Test a 64 bit integer data type."""
    buffer = bytearray([0xFF, 0xFF, 0xFF, 0xFF, 0xF8, 0xA4, 0x32, 0xEB])
    data_type = data_types.Int64.from_bytes(buffer)
    assert data_type.value == -1498954336607141889
    assert data_type.size == 8
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "Int64(value=-1498954336607141889)"
    assert repr(data_types.Int64()) == "Int64()"
    assert data_type == data_types.Int64.from_bytes(buffer)
    assert data_type == -1498954336607141889


def test_uint64() -> None:
    """Test a unsigned 64 bit integer data type."""
    buffer = bytearray([0x45, 0x49, 0x50, 0x51, 0x52, 0x53, 0x54, 0x55])
    data_type = data_types.UInt64.from_bytes(buffer)
    assert data_type.value == 6148631004284209477
    assert data_type.size == 8
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "UInt64(value=6148631004284209477)"
    assert repr(data_types.UInt64()) == "UInt64()"
    assert data_type == data_types.UInt64.from_bytes(buffer)
    assert data_type == 6148631004284209477


def test_ipv4() -> None:
    """Test an IPv4 data type."""
    buffer = bytearray([0x7F, 0x00, 0x00, 0x01])
    data_type = data_types.IPv4.from_bytes(buffer)
    assert data_type.value == "127.0.0.1"
    assert data_type.size == 4
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "IPv4(value=127.0.0.1)"
    assert repr(data_types.IPv4()) == "IPv4()"
    assert data_type == data_types.IPv4.from_bytes(buffer)
    assert data_type == "127.0.0.1"


def test_ipv6() -> None:
    """Test an IPv6 data type."""
    buffer = bytearray(
        [
            0xFE,
            0xED,
            0xDE,
            0xAD,
            0xBE,
            0xEF,
            0x0,
            0x0,
            0x0,
            0x0,
            0x0,
            0x0,
            0x0,
            0x0,
            0x0,
            0x1,
        ]
    )

    data_type = data_types.IPv6.from_bytes(buffer)
    assert data_type.value == "feed:dead:beef::1"
    assert data_type.size == 16
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "IPv6(value=feed:dead:beef::1)"
    assert repr(data_types.IPv6()) == "IPv6()"
    assert data_type == data_types.IPv6.from_bytes(buffer)
    assert data_type == "feed:dead:beef::1"


def test_string() -> None:
    """Test a string data type."""
    buffer = b"test\x00"
    data_type = data_types.String.from_bytes(buffer)
    assert data_type.value == "test"
    assert data_type.size == 5
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "String(value=test)"
    assert repr(data_types.String()) == "String(value=)"
    assert data_type == data_types.String.from_bytes(buffer)
    assert data_type == "test"

    # Test with unknown unicode char.
    buffer = b"test\xd8\x00"
    data_type = data_types.String.from_bytes(buffer)
    assert data_type.value == "test�"
    assert data_type.size == 6


def test_var_bytes() -> None:
    """Test a variable bytes data type."""
    buffer = b"\x04\xde\xad\xbe\xef"
    data_type = data_types.VarBytes.from_bytes(buffer)
    assert data_type.value == bytearray([0xDE, 0xAD, 0xBE, 0xEF])
    assert data_type.size == 5
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "VarBytes(value=b'\\xde\\xad\\xbe\\xef')"
    assert repr(data_types.VarBytes()) == "VarBytes(value=b'')"
    assert data_type == data_types.VarBytes.from_bytes(buffer)
    assert data_type == bytearray([0xDE, 0xAD, 0xBE, 0xEF])


def test_var_string() -> None:
    """Test a variable string data type."""
    buffer = b"\x04test"
    data_type = data_types.VarString.from_bytes(buffer)
    assert data_type.value == "test"
    assert data_type.size == 5
    assert data_type.to_bytes() == buffer
    assert repr(data_type) == "VarString(value=test)"
    assert repr(data_types.VarString()) == "VarString(value=)"
    assert data_type == data_types.VarString.from_bytes(buffer)
    assert data_type == "test"



"""Contains protocol representation."""

from __future__ import annotations

from abc import ABC, abstractmethod
import asyncio
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
import logging

from typing_extensions import TypeAlias

from pyplumio.const import ATTR_CONNECTED, DeviceType
from pyplumio.devices import PhysicalDevice
from pyplumio.exceptions import ProtocolError
from pyplumio.frames import Frame
from pyplumio.frames.requests import StartMasterRequest
from pyplumio.helpers.event_manager import EventManager
from pyplumio.stream import FrameReader, FrameWriter
from pyplumio.structures.network_info import (
    EthernetParameters,
    NetworkInfo,
    WirelessParameters,
)

_LOGGER = logging.getLogger(__name__)

_Callback: TypeAlias = Callable[[], Awaitable[None]]


class Protocol(ABC):
    """Represents a protocol."""

    connected: asyncio.Event
    reader: FrameReader | None
    writer: FrameWriter | None
    _on_connection_lost: set[_Callback]

    def __init__(self) -> None:
        """Initialize a new protocol."""
        super().__init__()
        self.connected = asyncio.Event()
        self.reader = None
        self.writer = None
        self._on_connection_lost = set()

    async def close_writer(self) -> None:
        """Ensure that writer is closed."""
        if self.writer:
            await self.writer.close()
            self.writer = None

    @property
    def on_connection_lost(self) -> set[_Callback]:
        """Return the callbacks that'll be called on connection lost."""
        return self._on_connection_lost

    @abstractmethod
    def connection_established(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Do something on connection established."""

    @abstractmethod
    async def connection_lost(self) -> None:
        """Do something on connection lost."""

    @abstractmethod
    async def shutdown(self) -> None:
        """Shutdown the protocol."""


class DummyProtocol(Protocol):
    """Represents a dummy protocol.

    This protocol sets frame reader and writer as attributes, then
    sets connected event and does nothing.
    """

    def connection_established(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Set reader and writer attributes and set the connected event."""
        self.reader = FrameReader(reader)
        self.writer = FrameWriter(writer)
        self.connected.set()

    async def connection_lost(self) -> None:
        """Close writer and call connection lost callbacks."""
        if self.connected.is_set():
            self.connected.clear()
            await self.close_writer()
            await asyncio.gather(*(callback() for callback in self.on_connection_lost))

    async def shutdown(self) -> None:
        """Shutdown the protocol."""
        if self.connected.is_set():
            self.connected.clear()
            await self.close_writer()


@dataclass
class Queues:
    """Represents asyncio queues."""

    __slots__ = ("read", "write")

    read: asyncio.Queue[Frame]
    write: asyncio.Queue[Frame]

    async def join(self) -> None:
        """Wait for queues to finish."""
        await asyncio.gather(self.read.join(), self.write.join())


class AsyncProtocol(Protocol, EventManager[PhysicalDevice]):
    """Represents an async protocol.

    This protocol implements producer-consumers pattern using
    asyncio queues.

    The frame producer tries to read frames from the write queue.
    If any is available, it sends them to the device via frame writer.

    It then reads stream via frame reader and puts received frame
    into the read queue.

    Frame consumers read frames from the read queue, create device
    entry, if needed, and send frame to the entry for the processing.
    """

    consumers_count: int
    _network: NetworkInfo
    _queues: Queues

    def __init__(
        self,
        ethernet_parameters: EthernetParameters | None = None,
        wireless_parameters: WirelessParameters | None = None,
        consumers_count: int = 3,
    ) -> None:
        """Initialize a new async protocol."""
        super().__init__()
        self.consumers_count = consumers_count
        self._network = NetworkInfo(
            eth=ethernet_parameters or EthernetParameters(status=False),
            wlan=wireless_parameters or WirelessParameters(status=False),
        )
        self._queues = Queues(read=asyncio.Queue(), write=asyncio.Queue())

    def connection_established(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        """Start frame producer and consumers."""
        self.reader = FrameReader(reader)
        self.writer = FrameWriter(writer)
        self._queues.write.put_nowait(StartMasterRequest(recipient=DeviceType.ECOMAX))
        self.create_task(
            self.frame_producer(self._queues, reader=self.reader, writer=self.writer),
            name="frame_producer_task",
        )
        for consumer in range(self.consumers_count):
            self.create_task(
                self.frame_consumer(self._queues.read),
                name=f"frame_consumer_task ({consumer})",
            )

        for device in self.data.values():
            device.dispatch_nowait(ATTR_CONNECTED, True)

        self.connected.set()

    async def _connection_close(self) -> None:
        """Close the connection if it is established."""
        self.connected.clear()
        await asyncio.gather(
            *(device.dispatch(ATTR_CONNECTED, False) for device in self.data.values())
        )
        await self.close_writer()

    async def connection_lost(self) -> None:
        """Close the connection and call connection lost callbacks."""
        if self.connected.is_set():
            await self._connection_close()
            await asyncio.gather(*(callback() for callback in self.on_connection_lost))

    async def shutdown(self) -> None:
        """Shutdown the protocol and close the connection."""
        await self._queues.join()
        self.cancel_tasks()
        await self.wait_until_done()
        if self.connected.is_set():
            await self._connection_close()
            await asyncio.gather(*(device.shutdown() for device in self.data.values()))

    async def frame_producer(
        self, queues: Queues, reader: FrameReader, writer: FrameWriter
    ) -> None:
        """Handle frame reads and writes."""
        await self.connected.wait()
        while self.connected.is_set():
            try:
                if not queues.write.empty():
                    await writer.write(await queues.write.get())
                    queues.write.task_done()

                if response := await reader.read():
                    queues.read.put_nowait(response)

            except ProtocolError as e:
                _LOGGER.debug("Can't process received frame: %s", e)
            except (OSError, asyncio.TimeoutError):
                self.create_task(self.connection_lost())
                break
            except Exception:
                _LOGGER.exception("Unexpected exception")

    async def frame_consumer(self, queue: asyncio.Queue[Frame]) -> None:
        """Handle frame processing."""
        await self.connected.wait()
        while self.connected.is_set():
            frame = await queue.get()
            device = await self.get_device_entry(frame.sender)
            device.handle_frame(frame)
            queue.task_done()

    async def get_device_entry(self, device_type: DeviceType) -> PhysicalDevice:
        """Set up or return a device entry."""
        name = device_type.name.lower()
        if name not in self.data:
            device = await PhysicalDevice.create(
                device_type, queue=self._queues.write, network=self._network
            )
            device.dispatch_nowait(ATTR_CONNECTED, True)
            self.create_task(device.async_setup(), name=f"device_setup_task ({name})")
            await self.dispatch(name, device)

        return self.data[name]



"""Contains tests for the main module."""

from unittest.mock import AsyncMock, call, patch

with patch("asyncio.run") as mock_asyncio_run:
    from pyplumio.__main__ import main


async def test_main(capsys) -> None:
    """Test main."""
    mock_device = AsyncMock()
    mock_device.get.side_effect = ("one", "two")
    mock_connection = AsyncMock()
    mock_connection.__aenter__.return_value = mock_connection
    mock_connection.get.return_value = mock_device
    main_fn = mock_asyncio_run.call_args[0][0]

    with patch(
        "pyplumio.open_serial_connection", return_value=mock_connection
    ) as mock_serial:
        await main_fn

    assert main.__name__ == main_fn.__name__
    mock_asyncio_run.assert_called_once()
    mock_serial.assert_called_with("/dev/ttyUSB0", 115200)
    mock_connection.__aenter__.assert_awaited_once()
    mock_connection.get.assert_awaited_once_with("ecomax")
    calls = [call("sensors"), call("parameters")]
    mock_device.get.assert_has_calls(calls)
    captured = capsys.readouterr()
    assert captured.out == "one\ntwo\n"



from importlib.metadata import version as pkg_version
import os
import sys

sys.path.insert(0, os.path.abspath("../../"))


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "PyPlumIO"
copyright = "2023, Denis Paavilainen"
author = "Denis Paavilainen"
release = pkg_version("pyplumio")
version = ".".join(release.split(".")[:3])

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.githubpages",
    "sphinx.ext.autosectionlabel",
    "sphinx_rtd_theme",
    "sphinx_search.extension",
]
autosummary_generate = True

templates_path = ["_templates"]
exclude_patterns: list[str] = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "navigation_depth": 2,
}

# Suppress warnings about duplicate label.

suppress_warnings = ['autosectionlabel.*']



"""Contains various utility methods."""

from __future__ import annotations

from collections.abc import Mapping
from typing import TypeVar


def to_camelcase(text: str, overrides: Mapping[str, str] | None = None) -> str:
    """Convert snake_case to CamelCase."""
    if not overrides:
        return "".join((x.capitalize() or "_") for x in text.split("_"))

    return "".join(
        (x.capitalize() or "_") if x.lower() not in overrides else overrides[x.lower()]
        for x in text.split("_")
    )


KT = TypeVar("KT")  # Key type.
VT = TypeVar("VT")  # Value type.


def ensure_dict(initial: dict[KT, VT] | None, *args: dict[KT, VT]) -> dict[KT, VT]:
    """Create or merge multiple dictionaries."""
    data = initial if initial is not None else {}
    for extra in args:
        data |= extra

    return data



"""Contains a timeout decorator."""

from __future__ import annotations

import asyncio
from collections.abc import Awaitable, Callable, Coroutine
from functools import wraps
from typing import Any, TypeVar

from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")


def timeout(
    seconds: float,
) -> Callable[[Callable[P, Awaitable[T]]], Callable[P, Coroutine[Any, Any, T]]]:
    """Decorate a timeout for the awaitable."""

    def decorator(
        func: Callable[P, Awaitable[T]],
    ) -> Callable[P, Coroutine[Any, Any, T]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            return await asyncio.wait_for(func(*args, **kwargs), timeout=seconds)

        return wrapper

    return decorator



"""Contains an ecoMAX class."""

from __future__ import annotations

import asyncio
from collections.abc import Coroutine, Generator, Iterable, Sequence
import logging
import time
from typing import Any, Final

from pyplumio.const import (
    ATTR_PASSWORD,
    ATTR_SENSORS,
    ATTR_STATE,
    DeviceState,
    DeviceType,
    FrameType,
)
from pyplumio.devices import PhysicalDevice
from pyplumio.devices.mixer import Mixer
from pyplumio.devices.thermostat import Thermostat
from pyplumio.filters import on_change
from pyplumio.frames import DataFrameDescription, Frame, Request
from pyplumio.helpers.parameter import STATE_OFF, STATE_ON, ParameterValues, State
from pyplumio.helpers.schedule import Schedule, ScheduleDay
from pyplumio.structures.alerts import ATTR_TOTAL_ALERTS
from pyplumio.structures.ecomax_parameters import (
    ATTR_ECOMAX_CONTROL,
    ATTR_ECOMAX_PARAMETERS,
    ECOMAX_CONTROL_PARAMETER,
    ECOMAX_PARAMETERS,
    THERMOSTAT_PROFILE_PARAMETER,
    EcomaxNumber,
    EcomaxSwitch,
    EcomaxSwitchDescription,
)
from pyplumio.structures.fuel_consumption import ATTR_FUEL_CONSUMPTION
from pyplumio.structures.mixer_parameters import ATTR_MIXER_PARAMETERS
from pyplumio.structures.mixer_sensors import ATTR_MIXER_SENSORS
from pyplumio.structures.network_info import ATTR_NETWORK, NetworkInfo
from pyplumio.structures.product_info import ATTR_PRODUCT, ProductInfo
from pyplumio.structures.regulator_data_schema import ATTR_REGDATA_SCHEMA
from pyplumio.structures.schedules import (
    ATTR_SCHEDULE_PARAMETERS,
    ATTR_SCHEDULES,
    SCHEDULE_PARAMETERS,
    SCHEDULES,
    ScheduleNumber,
    ScheduleSwitch,
    ScheduleSwitchDescription,
)
from pyplumio.structures.thermostat_parameters import (
    ATTR_THERMOSTAT_PARAMETERS,
    ATTR_THERMOSTAT_PROFILE,
)
from pyplumio.structures.thermostat_sensors import ATTR_THERMOSTAT_SENSORS

ATTR_MIXERS: Final = "mixers"
ATTR_THERMOSTATS: Final = "thermostats"
ATTR_FUEL_BURNED: Final = "fuel_burned"

NANOSECONDS_IN_SECOND: Final = 1_000_000_000
MAX_TIME_SINCE_LAST_FUEL_UPDATE_NS: Final = 300 * NANOSECONDS_IN_SECOND

SETUP_FRAME_TYPES: tuple[DataFrameDescription, ...] = (
    DataFrameDescription(
        frame_type=FrameType.REQUEST_UID,
        provides=ATTR_PRODUCT,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_REGULATOR_DATA_SCHEMA,
        provides=ATTR_REGDATA_SCHEMA,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_ECOMAX_PARAMETERS,
        provides=ATTR_ECOMAX_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_ALERTS,
        provides=ATTR_TOTAL_ALERTS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_SCHEDULES,
        provides=ATTR_SCHEDULES,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_MIXER_PARAMETERS,
        provides=ATTR_MIXER_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_THERMOSTAT_PARAMETERS,
        provides=ATTR_THERMOSTAT_PARAMETERS,
    ),
    DataFrameDescription(
        frame_type=FrameType.REQUEST_PASSWORD,
        provides=ATTR_PASSWORD,
    ),
)

_LOGGER = logging.getLogger(__name__)


class EcoMAX(PhysicalDevice):
    """Represents an ecoMAX controller."""

    address = DeviceType.ECOMAX
    _setup_frames = SETUP_FRAME_TYPES

    _fuel_burned_time_ns: int

    def __init__(self, queue: asyncio.Queue[Frame], network: NetworkInfo) -> None:
        """Initialize a new ecoMAX controller."""
        super().__init__(queue, network)
        self.subscribe(ATTR_ECOMAX_PARAMETERS, self._handle_ecomax_parameters)
        self.subscribe(ATTR_FUEL_CONSUMPTION, self._add_burned_fuel_meter)
        self.subscribe(ATTR_MIXER_PARAMETERS, self._handle_mixer_parameters)
        self.subscribe(ATTR_MIXER_SENSORS, self._handle_mixer_sensors)
        self.subscribe(ATTR_SCHEDULES, self._add_schedules)
        self.subscribe(ATTR_SCHEDULE_PARAMETERS, self._add_schedule_parameters)
        self.subscribe(ATTR_SENSORS, self._handle_ecomax_sensors)
        self.subscribe(ATTR_STATE, on_change(self._add_ecomax_control_parameter))
        self.subscribe(ATTR_THERMOSTAT_PARAMETERS, self._handle_thermostat_parameters)
        self.subscribe(ATTR_THERMOSTAT_PROFILE, self._add_thermostat_profile_parameter)
        self.subscribe(ATTR_THERMOSTAT_SENSORS, self._handle_thermostat_sensors)
        self._fuel_burned_time_ns = time.perf_counter_ns()

    async def async_setup(self) -> bool:
        """Set up an ecoMAX controller."""
        await self.wait_for(ATTR_SENSORS)
        return await super().async_setup()

    def handle_frame(self, frame: Frame) -> None:
        """Handle frame received from the ecoMAX device."""
        if isinstance(frame, Request) and (
            response := frame.response(data={ATTR_NETWORK: self._network})
        ):
            self.queue.put_nowait(response)

        super().handle_frame(frame)

    def _mixers(self, indexes: Iterable[int]) -> Generator[Mixer, None, None]:
        """Iterate through the mixer indexes.

        For each index, return or create an instance of the mixer class.
        Once done, dispatch the 'mixers' event without waiting.
        """
        mixers: dict[int, Mixer] = self.data.setdefault(ATTR_MIXERS, {})
        for index in indexes:
            yield mixers.setdefault(index, Mixer(self.queue, parent=self, index=index))

        return self.dispatch_nowait(ATTR_MIXERS, mixers)

    def _thermostats(self, indexes: Iterable[int]) -> Generator[Thermostat, None, None]:
        """Iterate through the thermostat indexes.

        For each index, return or create an instance of the thermostat
        class. Once done, dispatch the 'thermostats' event without
        waiting.
        """
        thermostats: dict[int, Thermostat] = self.data.setdefault(ATTR_THERMOSTATS, {})
        for index in indexes:
            yield thermostats.setdefault(
                index, Thermostat(self.queue, parent=self, index=index)
            )

        return self.dispatch_nowait(ATTR_THERMOSTATS, thermostats)

    async def _handle_ecomax_parameters(
        self, parameters: Sequence[tuple[int, ParameterValues]]
    ) -> bool:
        """Handle ecoMAX parameters.

        For each parameter dispatch an event with the parameter's name
        and value.
        """
        product: ProductInfo = await self.get(ATTR_PRODUCT)

        def _ecomax_parameter_events() -> Generator[Coroutine, Any, None]:
            """Get dispatch calls for ecoMAX parameter events."""
            for index, values in parameters:
                try:
                    description = ECOMAX_PARAMETERS[product.type][index]
                except IndexError:
                    _LOGGER.warning(
                        "Encountered unknown ecoMAX parameter (%i): %s. "
                        "Your device isn't fully compatible with this software "
                        "and may not work properly. Please visit the issue tracker "
                        "and open a feature request to support %s",
                        index,
                        values,
                        product.model,
                    )
                    return

                handler = (
                    EcomaxSwitch
                    if isinstance(description, EcomaxSwitchDescription)
                    else EcomaxNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self, description=description, values=values, index=index
                    ),
                )

        await asyncio.gather(*_ecomax_parameter_events())
        return True

    async def _add_burned_fuel_meter(self, fuel_consumption: float) -> None:
        """Calculate and dispatch the amount of fuel burned.

        This method calculates the fuel burned based on the time
        elapsed since the last sensor message, which contains fuel
        consumption data. If the elapsed time is within the acceptable
        range, it dispatches the fuel burned data. Otherwise, it logs a
        warning and skips the outdated data.
        """
        time_ns = time.perf_counter_ns()
        nanoseconds_passed = time_ns - self._fuel_burned_time_ns
        self._fuel_burned_time_ns = time_ns
        if nanoseconds_passed < MAX_TIME_SINCE_LAST_FUEL_UPDATE_NS:
            return self.dispatch_nowait(
                ATTR_FUEL_BURNED,
                fuel_consumption * nanoseconds_passed / (3600 * NANOSECONDS_IN_SECOND),
            )

        _LOGGER.warning(
            "Skipping outdated fuel consumption data: %f (was %i seconds old)",
            fuel_consumption,
            nanoseconds_passed / NANOSECONDS_IN_SECOND,
        )

    async def _handle_mixer_parameters(
        self,
        parameters: dict[int, Sequence[tuple[int, ParameterValues]]] | None,
    ) -> bool:
        """Handle mixer parameters.

        For each parameter dispatch an event with the
        parameter's name and value. Events are dispatched for the
        respective mixer instance.
        """
        if parameters:
            await asyncio.gather(
                *(
                    mixer.dispatch(ATTR_MIXER_PARAMETERS, parameters[mixer.index])
                    for mixer in self._mixers(indexes=parameters.keys())
                )
            )
            return True

        return False

    async def _handle_mixer_sensors(
        self, sensors: dict[int, dict[str, Any]] | None
    ) -> bool:
        """Handle mixer sensors.

        For each sensor dispatch an event with the
        sensor's name and value. Events are dispatched for the
        respective mixer instance.
        """
        if sensors:
            await asyncio.gather(
                *(
                    mixer.dispatch(ATTR_MIXER_SENSORS, sensors[mixer.index])
                    for mixer in self._mixers(indexes=sensors.keys())
                )
            )
            return True

        return False

    async def _add_schedules(
        self, schedules: list[tuple[int, list[list[bool]]]]
    ) -> dict[str, Schedule]:
        """Add schedules to the dataset."""
        return {
            SCHEDULES[index]: Schedule(
                name=SCHEDULES[index],
                device=self,
                monday=ScheduleDay.from_iterable(schedule[1]),
                tuesday=ScheduleDay.from_iterable(schedule[2]),
                wednesday=ScheduleDay.from_iterable(schedule[3]),
                thursday=ScheduleDay.from_iterable(schedule[4]),
                friday=ScheduleDay.from_iterable(schedule[5]),
                saturday=ScheduleDay.from_iterable(schedule[6]),
                sunday=ScheduleDay.from_iterable(schedule[0]),
            )
            for index, schedule in schedules
        }

    async def _add_schedule_parameters(
        self, parameters: Sequence[tuple[int, ParameterValues]]
    ) -> bool:
        """Add schedule parameters to the dataset."""

        def _schedule_parameter_events() -> Generator[Coroutine, Any, None]:
            """Get dispatch calls for schedule parameter events."""
            for index, values in parameters:
                description = SCHEDULE_PARAMETERS[index]
                handler = (
                    ScheduleSwitch
                    if isinstance(description, ScheduleSwitchDescription)
                    else ScheduleNumber
                )
                yield self.dispatch(
                    description.name,
                    handler.create_or_update(
                        device=self, description=description, values=values, index=index
                    ),
                )

        await asyncio.gather(*_schedule_parameter_events())
        return True

    async def _handle_ecomax_sensors(self, sensors: dict[str, Any]) -> bool:
        """Handle ecoMAX sensors.

        For each sensor dispatch an event with the sensor's name and
        value.
        """
        await asyncio.gather(
            *(self.dispatch(name, value) for name, value in sensors.items())
        )
        return True

    async def _add_ecomax_control_parameter(self, mode: DeviceState) -> None:
        """Create ecoMAX control parameter instance and dispatch an event."""
        await self.dispatch(
            ECOMAX_CONTROL_PARAMETER.name,
            EcomaxSwitch.create_or_update(
                description=ECOMAX_CONTROL_PARAMETER,
                device=self,
                values=ParameterValues(
                    value=int(mode != DeviceState.OFF), min_value=0, max_value=1
                ),
            ),
        )

    async def _handle_thermostat_parameters(
        self,
        parameters: dict[int, Sequence[tuple[int, ParameterValues]]] | None,
    ) -> bool:
        """Handle thermostat parameters.

        For each parameter dispatch an event with the
        parameter's name and value. Events are dispatched for the
        respective thermostat instance.
        """
        if parameters:
            await asyncio.gather(
                *(
                    thermostat.dispatch(
                        ATTR_THERMOSTAT_PARAMETERS, parameters[thermostat.index]
                    )
                    for thermostat in self._thermostats(indexes=parameters.keys())
                )
            )
            return True

        return False

    async def _add_thermostat_profile_parameter(
        self, values: ParameterValues | None
    ) -> EcomaxNumber | None:
        """Add thermostat profile parameter to the dataset."""
        if values:
            return EcomaxNumber(
                device=self, description=THERMOSTAT_PROFILE_PARAMETER, values=values
            )

        return None

    async def _handle_thermostat_sensors(
        self, sensors: dict[int, dict[str, Any]] | None
    ) -> bool:
        """Handle thermostat sensors.

        For each sensor dispatch an event with the
        sensor's name and value. Events are dispatched for the
        respective thermostat instance.
        """
        if sensors:
            await asyncio.gather(
                *(
                    thermostat.dispatch(
                        ATTR_THERMOSTAT_SENSORS, sensors[thermostat.index]
                    )
                    for thermostat in self._thermostats(indexes=sensors.keys())
                ),
                return_exceptions=True,
            )
            return True

        return False

    async def _set_ecomax_state(self, state: State) -> bool:
        """Try to set the ecoMAX control state."""
        try:
            switch: EcomaxSwitch = self.data[ATTR_ECOMAX_CONTROL]
            return await switch.set(state)
        except KeyError:
            _LOGGER.error("ecoMAX control is not available. Please try again later.")

        return False

    async def turn_on(self) -> bool:
        """Turn on the ecoMAX controller."""
        return await self._set_ecomax_state(STATE_ON)

    async def turn_off(self) -> bool:
        """Turn off the ecoMAX controller."""
        return await self._set_ecomax_state(STATE_OFF)

    def turn_on_nowait(self) -> None:
        """Turn on the ecoMAX controller without waiting."""
        self.create_task(self.turn_on())

    def turn_off_nowait(self) -> None:
        """Turn off the ecoMAX controller without waiting."""
        self.create_task(self.turn_off())

    async def shutdown(self) -> None:
        """Shutdown tasks for the ecoMAX controller and sub-devices."""
        mixers: dict[str, Mixer] = self.get_nowait(ATTR_MIXERS, {})
        thermostats: dict[str, Thermostat] = self.get_nowait(ATTR_THERMOSTATS, {})
        devices = (mixers | thermostats).values()
        await asyncio.gather(*(device.shutdown() for device in devices))
        await super().shutdown()



"""Contains a product info structure decoder."""

from __future__ import annotations

from dataclasses import dataclass
from functools import cache
import re
import struct
from typing import Any, Final

from pyplumio.const import ProductType
from pyplumio.helpers.data_types import UnsignedShort, VarBytes, VarString
from pyplumio.helpers.uid import unpack_uid
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_PRODUCT: Final = "product"


@cache
def format_model_name(model_name: str) -> str:
    """Format a device model name."""
    if m := re.match(r"^([A-Z]+)\s{0,}([0-9]{3,})(.+)$", model_name, re.IGNORECASE):
        model_device, model_number, model_suffix = m.groups()
        model_device = "ecoMAX" if model_device == "EM" else model_device
        return f"{model_device} {model_number}{model_suffix}"

    return model_name


@dataclass
class ProductInfo:
    """Represents a product info provided by an UID response."""

    __slots__ = ("type", "id", "uid", "logo", "image", "model")

    type: ProductType
    id: int
    uid: str
    logo: int
    image: int
    model: str


class ProductInfoStructure(StructureDecoder):
    """Represents a product info data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        product_type, product_id = struct.unpack_from("<BH", message)
        offset += 3
        uid = VarBytes.from_bytes(message, offset)
        offset += uid.size
        logo = UnsignedShort.from_bytes(message, offset)
        offset += logo.size
        image = UnsignedShort.from_bytes(message, offset)
        offset += image.size
        model_name = VarString.from_bytes(message, offset)
        offset += model_name.size

        return (
            ensure_dict(
                data,
                {
                    ATTR_PRODUCT: ProductInfo(
                        type=ProductType(product_type),
                        id=product_id,
                        uid=unpack_uid(uid.value),
                        logo=logo.value,
                        image=image.value,
                        model=format_model_name(model_name.value),
                    )
                },
            ),
            offset,
        )



"""Contains UID helpers."""

from __future__ import annotations

from functools import reduce
from typing import Final

CRC: Final = 0xA3A3
POLYNOMIAL: Final = 0xA001
BASE5_KEY: Final = "0123456789ABCDEFGHIJKLMNZPQRSTUV"


def unpack_uid(buffer: bytes) -> str:
    """Unpack UID from bytes."""
    return _base5(buffer + _crc16(buffer))


def _base5(buffer: bytes) -> str:
    """Encode bytes to a base5 encoded string."""
    number = int.from_bytes(buffer, byteorder="little")
    output = []
    while number:
        output.append(BASE5_KEY[number & 0b00011111])
        number >>= 5

    return "".join(reversed(output))


def _crc16(buffer: bytes) -> bytes:
    """Return a CRC 16."""
    crc16 = reduce(_crc16_byte, buffer, CRC)
    return crc16.to_bytes(length=2, byteorder="little")


def _crc16_byte(crc: int, byte: int) -> int:
    """Add a byte to the CRC."""
    crc ^= byte
    for _ in range(8):
        crc = (crc >> 1) ^ POLYNOMIAL if crc & 1 else crc >> 1

    return crc



"""Contains a network info structure decoder."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Final

from pyplumio.const import EncryptionType
from pyplumio.helpers.data_types import IPv4, VarString
from pyplumio.structures import Structure
from pyplumio.utils import ensure_dict

ATTR_NETWORK: Final = "network"

DEFAULT_IP: Final = "0.0.0.0"
DEFAULT_NETMASK: Final = "255.255.255.0"

NETWORK_INFO_SIZE: Final = 25


@dataclass(frozen=True)
class EthernetParameters:
    """Represents an ethernet parameters."""

    #: IP address
    ip: str = DEFAULT_IP

    #: IP subnet mask
    netmask: str = DEFAULT_NETMASK

    #: Gateway IP address
    gateway: str = DEFAULT_IP

    #: Connection status. Parameters will be ignored if set to `False`
    status: bool = True


@dataclass(frozen=True)
class WirelessParameters(EthernetParameters):
    """Represents a wireless network parameters."""

    #: Wireless Service Set IDentifier
    ssid: str = ""

    #: Wireless encryption standard
    #: (0 - unknown, 1 - no encryption, 2 - WEP, 3 - WPA, 4 - WPA2)
    encryption: EncryptionType = EncryptionType.NONE

    #: Wireless signal strength in percentage
    signal_quality: int = 100


@dataclass(frozen=True)
class NetworkInfo:
    """Represents a network parameters."""

    eth: EthernetParameters = field(default_factory=EthernetParameters)
    wlan: WirelessParameters = field(default_factory=WirelessParameters)
    server_status: bool = True


class NetworkInfoStructure(Structure):
    """Represents a network info data structure."""

    __slots__ = ()

    def encode(self, data: dict[str, Any]) -> bytearray:
        """Encode data to the bytearray message."""
        network_info: NetworkInfo = data.get(ATTR_NETWORK, NetworkInfo())
        return bytearray(
            b"\1"
            + IPv4(network_info.eth.ip).to_bytes()
            + IPv4(network_info.eth.netmask).to_bytes()
            + IPv4(network_info.eth.gateway).to_bytes()
            + network_info.eth.status.to_bytes(length=1, byteorder="little")
            + IPv4(network_info.wlan.ip).to_bytes()
            + IPv4(network_info.wlan.netmask).to_bytes()
            + IPv4(network_info.wlan.gateway).to_bytes()
            + network_info.server_status.to_bytes(length=1, byteorder="little")
            + network_info.wlan.encryption.to_bytes(length=1, byteorder="little")
            + network_info.wlan.signal_quality.to_bytes(length=1, byteorder="little")
            + network_info.wlan.status.to_bytes(length=1, byteorder="little")
            + b"\0" * 4
            + VarString(network_info.wlan.ssid).to_bytes()
        )

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        return (
            ensure_dict(
                data,
                {
                    ATTR_NETWORK: NetworkInfo(
                        eth=EthernetParameters(
                            ip=IPv4.from_bytes(message, offset).value,
                            netmask=IPv4.from_bytes(message, offset + 4).value,
                            gateway=IPv4.from_bytes(message, offset + 8).value,
                            status=bool(message[offset + 13]),
                        ),
                        wlan=WirelessParameters(
                            ip=IPv4.from_bytes(message, offset + 13).value,
                            netmask=IPv4.from_bytes(message, offset + 17).value,
                            gateway=IPv4.from_bytes(message, offset + 21).value,
                            encryption=EncryptionType(int(message[offset + 26])),
                            signal_quality=int(message[offset + 27]),
                            status=bool(message[offset + 28]),
                            ssid=VarString.from_bytes(message, offset + 33).value,
                        ),
                        server_status=bool(message[offset + 25]),
                    )
                },
            ),
            offset + NETWORK_INFO_SIZE,
        )



"""Contains constants."""

from __future__ import annotations

from enum import Enum, IntEnum, unique
from typing import Any, Final, Literal

from typing_extensions import TypeAlias

# General attributes.
ATTR_CONNECTED: Final = "connected"
ATTR_COUNT: Final = "count"
ATTR_CURRENT_TEMP: Final = "current_temp"
ATTR_DEVICE_INDEX: Final = "device_index"
ATTR_FRAME_ERRORS: Final = "frame_errors"
ATTR_INDEX: Final = "index"
ATTR_LOADED: Final = "loaded"
ATTR_OFFSET: Final = "offset"
ATTR_PARAMETER: Final = "parameter"
ATTR_PASSWORD: Final = "password"
ATTR_SCHEDULE: Final = "schedule"
ATTR_SENSORS: Final = "sensors"
ATTR_START: Final = "start"
ATTR_STATE: Final = "state"
ATTR_SWITCH: Final = "switch"
ATTR_TARGET_TEMP: Final = "target_temp"
ATTR_THERMOSTAT: Final = "thermostat"
ATTR_TRANSMISSION: Final = "transmission"
ATTR_TYPE: Final = "type"
ATTR_VALUE: Final = "value"
ATTR_SIZE: Final = "size"

# Bytes.
BYTE_UNDEFINED: Final = 0xFF


@unique
class EncryptionType(IntEnum):
    """Contains wireless encryption types."""

    UNKNOWN = 0
    NONE = 1
    WEP = 2
    WPA = 3
    WPA2 = 4


@unique
class DeviceState(IntEnum):
    """Contains device states."""

    OFF = 0
    STABILIZATION = 1
    KINDLING = 2
    WORKING = 3
    SUPERVISION = 4
    PAUSED = 5
    STANDBY = 6
    BURNING_OFF = 7
    ALERT = 8
    MANUAL = 9
    UNSEALING = 10
    OTHER = 11

    @classmethod
    def _missing_(cls, value: Any) -> DeviceState | None:
        """Look up state in extra state table."""
        return EXTRA_DEVICE_STATES.get(value, None)


EXTRA_DEVICE_STATES: Final = {
    12: DeviceState.STABILIZATION,  # ecoMAX 810P-L TOUCH
    23: DeviceState.STABILIZATION,  # ecoMAX 860P3-O
}


@unique
class LambdaState(IntEnum):
    """Contain lambda sensor states."""

    STOP = 0
    START = 1
    WORKING = 3


@unique
class ProductType(IntEnum):
    """Contains product types."""

    ECOMAX_P = 0
    ECOMAX_I = 1


@unique
class AlertType(IntEnum):
    """Contains alert types."""

    POWER_LOSS = 0
    BOILER_TEMP_SENSOR_FAILURE = 1
    MAX_BOILER_TEMP_EXCEEDED = 2
    FEEDER_TEMP_SENSOR_FAILURE = 3
    MAX_FEEDER_TEMP_EXCEEDED = 4
    EXHAUST_TEMP_SENSOR_FAILURE = 5
    MAX_EXHAUST_TEMP_EXCEEDED = 6
    KINDLING_FAILURE = 7
    NO_FUEL = 8
    LEAK_DETECTED = 9
    PRESSURE_SENSOR_FAILURE = 10
    FAN_FAILURE = 11
    INSUFFICIENT_AIR_PRESSURE = 12
    BURN_OFF_FAILURE = 13
    FLAME_SENSOR_FAILURE = 14
    LINEAR_ACTUATOR_BLOCKED = 15
    INCORRECT_PARAMETERS = 16
    CONDENSATION_WARNING = 17
    BOILER_STB_TRIPPED = 18
    FEEDER_STB_TRIPPED = 19
    MIN_WATER_PRESSURE_EXCEEDED = 20
    MAX_WATER_PRESSURE_EXCEEDED = 21
    FEEDER_JAMMED = 22
    FLAMEOUT = 23
    EXHAUST_FAN_FAILURE = 24
    EXTERNAL_FEEDER_FAILURE = 25
    SOLAR_COLLECTOR_TEMP_SENSOR_FAILURE = 26
    SOLAR_CIRCUIT_TEMP_SENSOR_FAILURE = 27
    H1_CIRCUIT_TEMP_SENSOR_FAILURE = 28
    H2_CIRCUIT_TEMP_SENSOR_FAILURE = 29
    H3_CIRCUIT_TEMP_SENSOR_FAILURE = 30
    OUTDOOR_TEMP_SENSOR_FAILURE = 31
    WATER_HEATER_TEMP_SENSOR_FAILURE = 32
    H0_CIRCUIT_TEMP_SENSOR_FAILURE = 33
    FROST_PROTECTION_RUNNING_WO_HS = 34
    FROST_PROTECTION_RUNNING_W_HS = 35
    MAX_SOLAR_COLLECTOR_TEMP_EXCEEDED = 36
    MAX_HEATED_FLOOR_TEMP_EXCEEDED = 37
    BOILER_COOLING_RUNNING = 38
    ECOLAMBDA_CONNECTION_FAILURE = 39
    PRIMARY_AIR_THROTTLE_JAMMED = 40
    SECONDARY_AIR_THROTTLE_JAMMED = 41
    FEEDER_OVERFLOW = 42
    FURNANCE_OVERFLOW = 43
    MODULE_B_CONNECTION_FAILURE = 44
    CLEANING_ACTUATOR_FAILURE = 45
    MIN_PRESSURE_EXCEEDED = 46
    MAX_PRESSURE_EXCEEDED = 47
    PRESSURE_SENSOR_DAMAGED = 48
    MAX_MAIN_HS_TEMP_EXCEEDED = 49
    MAX_ADDITIONAL_HS_TEMP_EXCEEDED = 50
    SOLAR_PANEL_OFFLINE = 51
    FEEDER_CONTROL_FAILURE = 52
    FEEDER_BLOCKED = 53
    MAX_THERMOCOPLE_TEMP_EXCEEDED = 54
    THERMOCOUPLE_WIRING_FAILURE = 55
    UNKNOWN_ERROR = 255


@unique
class DeviceType(IntEnum):
    """Contains device addresses."""

    ALL = 0
    ECOMAX = 69
    ECOSTER = 81
    ECONET = 86


@unique
class FrameType(IntEnum):
    """Contains frame types."""

    REQUEST_STOP_MASTER = 24
    REQUEST_START_MASTER = 25
    REQUEST_CHECK_DEVICE = 48
    REQUEST_ECOMAX_PARAMETERS = 49
    REQUEST_MIXER_PARAMETERS = 50
    REQUEST_SET_ECOMAX_PARAMETER = 51
    REQUEST_SET_MIXER_PARAMETER = 52
    REQUEST_SCHEDULES = 54
    REQUEST_SET_SCHEDULE = 55
    REQUEST_UID = 57
    REQUEST_PASSWORD = 58
    REQUEST_ECOMAX_CONTROL = 59
    REQUEST_ALERTS = 61
    REQUEST_PROGRAM_VERSION = 64
    REQUEST_REGULATOR_DATA_SCHEMA = 85
    REQUEST_THERMOSTAT_PARAMETERS = 92
    REQUEST_SET_THERMOSTAT_PARAMETER = 93
    RESPONSE_DEVICE_AVAILABLE = 176
    RESPONSE_ECOMAX_PARAMETERS = 177
    RESPONSE_MIXER_PARAMETERS = 178
    RESPONSE_SET_ECOMAX_PARAMETER = 179
    RESPONSE_SET_MIXER_PARAMETER = 180
    RESPONSE_SCHEDULES = 182
    RESPONSE_UID = 185
    RESPONSE_PASSWORD = 186
    RESPONSE_ECOMAX_CONTROL = 187
    RESPONSE_ALERTS = 189
    RESPONSE_PROGRAM_VERSION = 192
    RESPONSE_REGULATOR_DATA_SCHEMA = 213
    RESPONSE_THERMOSTAT_PARAMETERS = 220
    RESPONSE_SET_THERMOSTAT_PARAMETER = 221
    MESSAGE_REGULATOR_DATA = 8
    MESSAGE_SENSOR_DATA = 53


@unique
class UnitOfMeasurement(Enum):
    """Contains units of measurement."""

    CELSIUS = "°C"
    KILO_WATT = "kW"
    GRAMS = "g"
    KILOGRAMS = "kg"
    KILO_WATT_HOUR_PER_KILOGRAM = "kWh/kg"
    KILOGRAMS_PER_HOUR = "kg/h"
    SECONDS = "s"
    MINUTES = "min"
    DAYS = "days"


PERCENTAGE: Final = "%"

STATE_ON: Final = "on"
STATE_OFF: Final = "off"


State: TypeAlias = Literal["on", "off"]



"""Contains data type helper classes."""

from __future__ import annotations

from abc import ABC, abstractmethod
import socket
import struct
from typing import ClassVar, Final, Generic, TypeVar

T = TypeVar("T")
DataTypeT = TypeVar("DataTypeT", bound="DataType")


class DataType(ABC, Generic[T]):
    """Represents a data type."""

    __slots__ = ("_value", "_size")

    _value: T
    _size: int

    def __init__(self, value: T | None = None) -> None:
        """Initialize a new data type."""
        if value is not None:
            self._value = value

        self._size = 0

    def __repr__(self) -> str:
        """Return serializable string representation of the class."""
        if hasattr(self, "_value"):
            return f"{self.__class__.__name__}(value={self._value})"

        return f"{self.__class__.__name__}()"

    def __eq__(self, other: object) -> bool:
        """Compare if this data type is equal to other."""
        if (
            isinstance(other, DataType)
            and hasattr(other, "_value")
            and hasattr(self, "_value")
        ):
            return bool(self._value == other._value)

        if (
            isinstance(other, DataType)
            and not hasattr(other, "_value")
            and not hasattr(self, "_value")
        ):
            return type(self) is type(other)

        if hasattr(self, "_value"):
            return bool(self._value == other)

        return NotImplemented

    def _slice_data(self, data: bytes) -> bytes:
        """Slice the data to data type size."""
        return data if self.size == 0 else data[: self.size]

    @classmethod
    def from_bytes(cls: type[DataTypeT], data: bytes, offset: int = 0) -> DataTypeT:
        """Initialize a new data type from bytes."""
        data_type = cls()
        data_type.unpack(data[offset:])
        return data_type

    def to_bytes(self) -> bytes:
        """Convert data type to bytes."""
        return self.pack()

    @property
    def value(self) -> T:
        """Return the data type value."""
        return self._value

    @property
    def size(self) -> int:
        """Return the data type size."""
        return self._size

    @abstractmethod
    def pack(self) -> bytes:
        """Pack the data."""

    @abstractmethod
    def unpack(self, data: bytes) -> None:
        """Unpack the data."""


class Undefined(DataType):
    """Represents an undefined."""

    __slots__ = ()

    def pack(self) -> bytes:
        """Pack the data."""
        return b""

    def unpack(self, _: bytes) -> None:
        """Unpack the data."""
        self._value = None


BITARRAY_LAST_INDEX: Final = 7


class BitArray(DataType[int]):
    """Represents a bit array."""

    __slots__ = ("_index",)

    _index: int

    def __init__(self, value: bool | None = None, index: int = 0) -> None:
        """Initialize a new bit array."""
        super().__init__(value)
        self._index = index

    def __repr__(self) -> str:
        """Return serializable string representation of the class."""
        if hasattr(self, "_value"):
            return (
                f"{self.__class__.__name__}(value={self._value}, index={self._index})"
            )

        return f"{self.__class__.__name__}(index={self._index})"

    def next(self, index: int = 0) -> int:
        """Set current bit and return the next index in the bit array."""
        self._index = index
        return 0 if self._index == BITARRAY_LAST_INDEX else self._index + 1

    def pack(self) -> bytes:
        """Pack the data."""
        if hasattr(self, "_value"):
            return UnsignedChar(self._value).to_bytes()

        return b""

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = UnsignedChar.from_bytes(data[:1]).value

    @property
    def value(self) -> bool:
        """Return the data type value."""
        if hasattr(self, "_value"):
            return bool(self._value & (1 << self._index))

        raise ValueError

    @property
    def size(self) -> int:
        """Return the data type size."""
        return 1 if self._index == BITARRAY_LAST_INDEX else 0


class IPv4(DataType[str]):
    """Represents an IPv4 address."""

    __slots__ = ()

    @property
    def size(self) -> int:
        """Return the data type size."""
        return 4

    def pack(self) -> bytes:
        """Pack the data."""
        return socket.inet_aton(self.value)

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = socket.inet_ntoa(self._slice_data(data))


class IPv6(DataType[str]):
    """Represents an IPv6 address."""

    __slots__ = ()

    @property
    def size(self) -> int:
        """Return a data type size."""
        return 16

    def pack(self) -> bytes:
        """Pack the data."""
        return socket.inet_pton(socket.AF_INET6, self.value)

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = socket.inet_ntop(socket.AF_INET6, self._slice_data(data))


class String(DataType[str]):
    """Represents a null-terminated string."""

    __slots__ = ()

    def __init__(self, value: str = "") -> None:
        """Initialize a new null-terminated string data type."""
        super().__init__(value)
        self._size = len(self.value) + 1

    def pack(self) -> bytes:
        """Pack the data."""
        return self.value.encode() + b"\0"

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = data.split(b"\0", 1)[0].decode("utf-8", "replace")
        self._size = len(self.value) + 1


class VarBytes(DataType[bytes]):
    """Represents a variable-length bytes."""

    __slots__ = ()

    def __init__(self, value: bytes = b"") -> None:
        """Initialize a new variable-length bytes data type."""
        super().__init__(value)
        self._size = len(value) + 1

    def pack(self) -> bytes:
        """Pack the data."""
        return UnsignedChar(self.size - 1).to_bytes() + self.value

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._size = data[0] + 1
        self._value = data[1 : self.size]


class VarString(DataType[str]):
    """Represents a variable length string."""

    __slots__ = ()

    def __init__(self, value: str = "") -> None:
        """Initialize a new variable length bytes data type."""
        super().__init__(value)
        self._size = len(value) + 1

    def pack(self) -> bytes:
        """Pack the data."""
        return UnsignedChar(self.size - 1).to_bytes() + self.value.encode()

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._size = data[0] + 1
        self._value = data[1 : self.size].decode("utf-8", "replace")


class BuiltInDataType(DataType[T], ABC):
    """Represents a data type that is supported by the struct module."""

    __slots__ = ("_struct",)

    _struct: ClassVar[struct.Struct]

    def pack(self) -> bytes:
        """Pack the data."""
        return self._struct.pack(self.value)

    def unpack(self, data: bytes) -> None:
        """Unpack the data."""
        self._value = self._struct.unpack_from(data)[0]

    @property
    def size(self) -> int:
        """Return a data type size."""
        if not self._size:
            self._size = self._struct.size

        return self._size


class SignedChar(BuiltInDataType[int]):
    """Represents a signed char."""

    __slots__ = ()

    _struct = struct.Struct("<b")


class UnsignedChar(BuiltInDataType[int]):
    """Represents an unsigned char."""

    __slots__ = ()

    _struct = struct.Struct("<B")


class Short(BuiltInDataType[int]):
    """Represents a 16-bit integer."""

    __slots__ = ()

    _struct = struct.Struct("<h")


class UnsignedShort(BuiltInDataType[int]):
    """Represents an unsigned 16-bit integer."""

    __slots__ = ()

    _struct = struct.Struct("<H")


class Int(BuiltInDataType[int]):
    """Represents a 32-bit integer."""

    __slots__ = ()

    _struct = struct.Struct("<i")


class UnsignedInt(BuiltInDataType[int]):
    """Represents a unsigned 32-bit integer."""

    __slots__ = ()

    _struct = struct.Struct("<I")


class Float(BuiltInDataType[float]):
    """Represents a float."""

    __slots__ = ()

    _struct = struct.Struct("<f")


class Double(BuiltInDataType[float]):
    """Represents a double."""

    __slots__ = ()

    _struct = struct.Struct("<d")


class Int64(BuiltInDataType[int]):
    """Represents a 64-bit signed integer."""

    __slots__ = ()

    _struct = struct.Struct("<q")


class UInt64(BuiltInDataType[int]):
    """Represents a 64-bit unsigned integer."""

    __slots__ = ()

    _struct = struct.Struct("<Q")


# The regdata type map links data type classes to their
# respective type ids in the regulator data schema.
DATA_TYPES: tuple[type[DataType], ...] = (
    Undefined,
    SignedChar,
    Short,
    Int,
    UnsignedChar,
    UnsignedShort,
    UnsignedInt,
    Float,
    Undefined,
    Double,
    BitArray,
    String,
    String,
    Int64,
    UInt64,
    IPv4,
    IPv6,
)



"""Contains a boiler load structure decoder."""

from __future__ import annotations

from typing import Any, Final

from pyplumio.const import BYTE_UNDEFINED
from pyplumio.structures import StructureDecoder
from pyplumio.utils import ensure_dict

ATTR_BOILER_LOAD: Final = "boiler_load"


class BoilerLoadStructure(StructureDecoder):
    """Represents a boiler load sensor data structure."""

    __slots__ = ()

    def decode(
        self, message: bytearray, offset: int = 0, data: dict[str, Any] | None = None
    ) -> tuple[dict[str, Any], int]:
        """Decode bytes and return message data and offset."""
        boiler_load = message[offset]
        offset += 1

        if boiler_load == BYTE_UNDEFINED:
            return ensure_dict(data), offset

        return (ensure_dict(data, {ATTR_BOILER_LOAD: boiler_load}), offset)
